// Generated by CoffeeScript 1.7.0
(function() {
  var Config, accord, coffee, fs, os, path, _;

  path = require('path');

  fs = require('fs');

  accord = require('accord');

  coffee = require('coffee-script');

  _ = require('lodash');

  os = require('os');


  /**
   * @class Config
   * @classdesc Manages configuration info and setup for roots
   */

  Config = (function() {

    /**
     * This is an escaped version of the platform specific path seperator to be
     * used in regular expressions. It is defined as a (derived) constant,
     * because the path seperator does not change over the course of a run.
     */
    var PATH_SEPERATOR_REGEXP_STRING, get_compilers, load_config;

    PATH_SEPERATOR_REGEXP_STRING = path.sep.replace('\\', '\\\\');


    /**
     * Creates a new instance of the roots config. This happens once, as soon as
     * you initialize the roots class. The constructor sets up a number of default
     * variables, explained below:
     *
     * - output: directory path appended to roots.root that results are written to
     * - dump_dirs: these directories dump their contents and appear transparent
     * - env: environment, usually development, staging, or production
     * - debug: when true, roots is more verbose in its output
     * - live_reload: when a compile finishes in watch more, reload the browser
     * - open_browser: when `roots watch` is run, open up the browser
     *
     * load_config loads in the app.coffee file, which can overwrite any of these
     * previous settings
     *
     * Then, we set up ignores if not provided by app.coffee, and add a couple
     * defaults that must always be ignored.
     *
     * Finally, we read the package.json file if present and grab all the adapters
     * needed to compile files in this project. More details on that later.
     *
     * @param  {Function} @roots - roots class instance
     *
     * @todo uniq filter ignores
     */

    function Config(roots, opts) {
      var _ref, _ref1, _ref2;
      this.roots = roots;
      this.output = 'public';
      this.dump_dirs = ['views', 'assets'];
      this.env = (_ref = opts.env) != null ? _ref : 'development';
      this.verbose = (_ref1 = opts.verbose) != null ? _ref1 : false;
      this.debug = false;
      this.live_reload = true;
      this.open_browser = (_ref2 = !opts.no_open) != null ? _ref2 : true;
      load_config.call(this);
      if (this.ignores == null) {
        this.ignores = [];
      }
      this.ignores = this.ignores.concat(['package.json', 'app*.coffee', "" + this.output + "/**/*", '.git']);
      if (this.watcher_ignores == null) {
        this.watcher_ignores = [];
      }
      this.watcher_ignores = this.watcher_ignores.concat(['package.json', 'app.coffee', 'node_modules/**/*', "" + this.output + "/**/*", '.git']);
      this.compilers = get_compilers.call(this);
    }


    /**
     * This function is responsible for loading the app.coffee file into the
     * config.
     *
     * First, it checks the environment. If it's 'development', the default, roots
     * loads 'app.coffee', and if there is a custom environment, it tries to load
     * 'app.ENV_NAME.coffee'. It then makes sure the config file exists. If not,
     * it just returns, and if there was a custom environment logs out a warning.
     *
     * If it does exist, there are two ways app.coffee can be processed. First is
     * 'simple mode', entered if it doesn't export anything when the file is
     * required. In this mode, the config file is processed as a coffeescript
     * object.
     *
     * If the file does export anything, this means it's being used as a node
     * file, so it is required and processed as a node file.
     *
     * Each of the values that are exported are attached directly to the config
     * object, overwriting the defaults if this applies. Finally, extensions are
     * all registered with roots if they are provided.
     */

    load_config = function() {
      var basename, conf, config_exists, config_path, k, v;
      basename = this.env === 'development' ? "app" : "app." + this.env;
      config_path = path.join(this.roots.root, basename);
      config_exists = fs.existsSync("" + config_path + ".coffee");
      if (!config_exists) {
        if (this.env !== 'development') {
          console.warn(("\nEnvironment config file not found. Make sure 'app." + this.env + ".coffee' is present at the root of your project.\n").yellow);
        }
        return;
      }
      conf = require(config_path);
      if (Object.keys(conf).length < 1) {
        conf = eval(coffee.compile(fs.readFileSync("" + config_path + ".coffee", 'utf8'), {
          bare: true
        }));
      }
      for (k in conf) {
        v = conf[k];
        this[k] = v;
      }
      if (this.extensions) {
        return this.roots.extensions.register(this.extensions);
      }
    };


    /**
     * Produces the full path to the output folder
     * @return {String} - path to output folder
     */

    Config.prototype.output_path = function() {
      return path.join(this.roots.root, this.output);
    };


    /**
     * Given a vinyl-wrapped file and optional extension, this function produces
     * the path to the file's destination. To do so, it goes through these steps:
     *
     * - Take the relative path and split it by /
     * - If it's in a 'dumped' directory, remove that directory
     * - URI encode any strange characters
     * - Add the full path to the output folder to the beginning
     * - Join it back together with /
     * - If an extension override was provided, replace the extension
     *
     * @param  {File} f - vinyl instance
     * @param  {String} ext - file extension, no dot
     * @return {String} path to where the file should be written
     */

    Config.prototype.out = function(f, ext) {
      var res;
      res = f.relative.split(path.sep);
      if (_.includes(this.dump_dirs, res[0])) {
        res.shift();
      }
      res.unshift(this.output_path());
      res = res.join(path.sep);
      if (ext) {
        res = res.replace(RegExp("\\.[^" + PATH_SEPERATOR_REGEXP_STRING + "]*$"), "." + ext);
      }
      return res;
    };


    /**
     * Grabs all adapters necessary to compile files in this project.
     * Scans the package.json file's dependencies for packages that have
     * registered accord adapters and loads those. Alerts if dependencies
     * have not been installed.
     *
     * @private
     *
     * @return {Array} - array of accord adapters
     */

    get_compilers = function() {
      var dep, err, local_compiler, pkg, pkg_json_path, res, _i, _len, _ref;
      res = {};
      pkg_json_path = path.join(this.roots.root, 'package.json');
      if (!fs.existsSync(pkg_json_path)) {
        return res;
      }
      pkg = require(pkg_json_path);
      _ref = _.keys(pkg.dependencies).concat(_.keys(pkg.devDependencies));
      for (_i = 0, _len = _ref.length; _i < _len; _i++) {
        dep = _ref[_i];
        if (accord.supports(dep)) {
          try {
            local_compiler = path.join(this.roots.root, 'node_modules', dep);
          } catch (_error) {
            err = _error;
            throw new Error("'" + dep + "' not found. install it with 'npm install'");
          }
          res[dep] = accord.load(dep, local_compiler);
        }
      }
      return res;
    };

    return Config;

  })();

  module.exports = Config;

}).call(this);
